import os

try:
    from engine.libs.hub import get_last_weight, get_best_weight
except ImportError:
    raise ImportError(
        'Please run "pip install granular-engine" to install engine')


def weight_and_experiment(url, best=False):
    if best:
        checkpoint, experiment_id = get_best_weight(url)
    else:
        checkpoint, experiment_id = get_last_weight(url)

    dst_path = 'pretrain/' + '/'.join(checkpoint.replace('gs://', '').replace('s3://', '').split('/')[2:])
    os.system(f"gsutil -m cp -n -r {checkpoint} {dst_path} 2> /dev/null")
            
    return dst_path, experiment_id




import base64

import torch
from torchvision.ops import nms
import numpy as np
import rasterio as rio
import cv2

from mmengine.config import Config
from mmengine.runner import Runner
from shapely.geometry import box
from mmdet.structures import DetDataSample


def rio2patches(img_r, cfg):
    """this function read a rio image and creates patches based on cfg.window and cfg.stride.
    return list of patches (patches x window_height x window_width) and list of location of patches (x1, y1, x2, y2).
    Make sure every pixel of image is included in patches. We cannot ignore left and bottom sides.
    Duplicates are okay.
    """
    window_height, window_width = cfg.window if "window" in cfg else (512, 512)
    stride_height, stride_width = cfg.stride if "stride" in cfg else (512, 512)

    img_height, img_width = img_r.shape[1], img_r.shape[2]

    if window_height > img_height:
        window_height = img_height

    if window_width > img_width:
        window_width = img_width

    patches = []
    locs = []

    for i in range(0, img_height, stride_height):
        for j in range(0, img_width, stride_width):
            if img_height < i + window_height:
                y_start = img_height - window_height
                y_end = img_height
            else:
                y_start = i
                y_end = i + window_height

            if img_width < j + window_width:
                x_start = img_width - window_width
                x_end = img_width
            else:
                x_start = j
                x_end = j + window_width

            patch = img_r[:, y_start:y_end, x_start:x_end]

            patches.append(patch)
            locs.append([x_start, y_start])

    return np.array(patches, dtype=np.float32), locs


def merge_seg_patches(mask_patches, location_of_patches):
    """use location of patches to merge logits of patches
    (patches x num_classes x window_height x window_width),
    in case of overlapping area use average of logits.
    After that apply softmax to the num_classes chanel.
    this should return a array of dimension (img_height, img_width)"""

    num_classes = mask_patches.shape[1]
    window_height, window_width = mask_patches.shape[2], mask_patches.shape[3]
    img_height, img_width = (
        location_of_patches[-1][1] + window_height,
        location_of_patches[-1][0] + window_width,
    )

    merged_logits = np.zeros((num_classes, img_height, img_width))
    overlap_count = np.zeros((num_classes, img_height, img_width))

    for i, patch in enumerate(mask_patches):
        x1, y1 = location_of_patches[i]
        x2, y2 = x1 + window_width, y1 + window_height
        merged_logits[:, y1:y2, x1:x2] += patch
        overlap_count[:, y1:y2, x1:x2] += 1

    averaged_merged_logits = merged_logits / overlap_count

    return np.argmax(averaged_merged_logits, axis=0)


def merge_bboxes(bboxes, scores, labels, location_of_patches):
    """This should be easy to merge, the bboxes locations will be localized
    to window_height and window_width, just translate them to original image
    location using location_of_patches. Once all bboxes are merged.
    Use apply non maximal supression, this will remove overlapping boxes area
    due to overlapping regions in patches."""

    flattened_bboxes = []
    flattened_scores = []
    flattened_labels = []

    for i, patch_bbox in enumerate(bboxes):
        x, y = location_of_patches[i]
        for bbox in patch_bbox:
            bbox[0] += x
            bbox[1] += y
            bbox[2] += x
            bbox[3] += y
            flattened_bboxes.append(bbox)
        flattened_scores += scores[i]
        flattened_labels += labels[i]

    keep_idx = nms(
        torch.tensor(flattened_bboxes),
        torch.tensor(flattened_scores),
        iou_threshold=0.5,
    )

    merged_bboxes = np.array(flattened_bboxes)[list(keep_idx.numpy())]
    merged_scores = np.array(flattened_scores)[list(keep_idx.numpy())]
    merged_labels = np.array(flattened_labels)[list(keep_idx.numpy())]

    return merged_bboxes, merged_scores, merged_labels


# def classification_as_object_detection(logit_patches, location_of_patches):
#     """This is similar to merge_seg_patches. In case of large images when we apply classification on them.
#        We will be doing that over patches but then we do need to present informatoion for whole image.
#        So all we do is we say that patch is a box. So we create boxes for each patches located in the large images
#        and assign the classification label to it. No need to worry about overlapping regions."""
#     pass


class Inference:
    def __init__(
        self,
        cfg_string,
        weight_file,
        GS_ACCESS_KEY_ID=None,
        GS_SECRET_ACCESS_KEY=None,
        logger=None,
    ):
        cfg = Config.fromstring(cfg_string, file_format=".py")
        cfg.launcher = "none"
        runner = Runner.from_cfg(cfg)

        checkpoint = torch.load(weight_file)
        runner.model.load_state_dict(checkpoint["state_dict"])
        _ = runner.model.eval()

        self.cfg = cfg
        self.runner = runner
        self.GS_ACCESS_KEY_ID = GS_ACCESS_KEY_ID
        self.GS_SECRET_ACCESS_KEY = GS_SECRET_ACCESS_KEY
        self.logger = logger

    def read_raster(self, tif_url):
        try:
            with rio.Env(
                GS_ACCESS_KEY_ID=self.GS_ACCESS_KEY_ID,
                GS_SECRET_ACCESS_KEY=self.GS_SECRET_ACCESS_KEY,
            ):
                r = rio.open(tif_url)
                img = r.read()
            self.logger.info(f"Reading raster {tif_url}")
            return r, img
        except:
            if self.logger:
                self.logger.error(f"Error in reading raster {tif_url}")

    def segmentation(self, tif_path):
        r, img = self.read_raster(tif_path)
        patches, location_of_patches = rio2patches(img, self.cfg)
        mask_patches = []

        if self.logger:
            self.logger.info(
                f"Total number of patches for image {tif_path} are: {len(patches)}"
            )

        for patch in patches:
            model_input = {"inputs": [torch.from_numpy(patch)]}
            if self.logger:
                self.logger.info(f"Input to the model: {model_input}")
            result = self.runner.model.test_step(model_input)
            if self.logger:
                self.logger.info(f"Result from the model: {result}")
            mask_patch = result[0].seg_logits.data.detach().cpu().numpy()
            if self.logger:
                self.logger.info(f"Mask: {mask_patch}")

            mask_patches.append(mask_patch)

        full_mask = merge_seg_patches(
            np.array(mask_patches), location_of_patches
        )

        if self.logger:
            self.logger.info(f"Full mask's dimensions are {full_mask.shape}")

        is_success, buffer = cv2.imencode(".png", full_mask)
        img_base64 = base64.b64encode(buffer).decode()
        lbl_ids = np.unique(full_mask)

        return img_base64, lbl_ids

    def detection(self, tif_path):
        r, img = self.read_raster(tif_path)
        patches, location_of_patches = rio2patches(img, self.cfg)
        img_shape = (patches.shape[2], patches.shape[3])

        if self.logger:
            self.logger.info(
                f"Total number of patches for image {tif_path} are: {len(patches)}"
            )

        bboxes, scores, labels = [], [], []
        for patch in patches:
            model_input = {
                "inputs": torch.from_numpy(patch).unsqueeze(0),
                "data_samples": [
                    DetDataSample(
                        metainfo={
                            "img_shape": img_shape,
                            "ori_shape": img_shape,
                            "scale_factor": (1, 1),
                        }
                    )
                ],
            }
            result = self.runner.model.test_step(model_input)[0]

            bboxes.append(
                list(result.pred_instances.bboxes.detach().cpu().numpy())
            )
            scores.append(
                list(result.pred_instances.scores.detach().cpu().numpy())
            )
            labels.append(
                list(result.pred_instances.labels.detach().cpu().numpy())
            )

        bboxes, scores, labels = merge_bboxes(
            bboxes, scores, labels, location_of_patches
        )

        geo_bboxes = []
        for bbox in bboxes:
            x1, y1, x2, y2 = bbox
            xs, ys = rio.transform.xy(r.transform, (y1, y2), (x1, x2))
            poly = box(xs[0], ys[1], xs[1], ys[0])
            geo_bboxes.append([list(poly.exterior.coords)])

        return geo_bboxes, scores, labels

    def classification(self, tif_path):
        pass



import json, os
from copy import deepcopy
import numpy as np
import pandas as pd
from tqdm import tqdm

from mmengine.registry import DATASETS
from typing import List, Union, Callable, Optional, Sequence

from shapely.geometry import box, Polygon, MultiPolygon

from mmengine.dataset import BaseDataset, Compose
from multiprocessing import Pool



@DATASETS.register_module()
class EngineCocoExt(BaseDataset):
    def __init__(self,
                 data_path: str,
                 raster_dir_path: str,
                 vector_dir_path: str,
                 split: str = "train",
                 class_title: str = None,
                 slice: bool = False,
                 window: tuple = None,
                 stride: tuple = None,
                 pipeline: List[Union[dict, Callable]] = [],
                 max_refetch: int = 1000,
                 ignore_index: int = 255,
                 reduce_zero_label: bool = False,
                 filter_empty: bool = False,
                 serialize_data: bool = True,
                 lazy_init: bool = False,
                 indices: Optional[Union[int, Sequence[int]]] = None,
                 test_mode: bool = False):
        self.raster_data_path = f"{data_path}/{raster_dir_path}"
        self.vector_data_path = f"{data_path}/{vector_dir_path}"

        self.split = split
        self.slice = slice 
        self.window = window 
        self.stride = stride
        self.class_title = class_title
        self.ignore_index = ignore_index
        self.reduce_zero_label = reduce_zero_label
        self.filter_empty = filter_empty
        self.serialize_data = serialize_data
        self._indices = indices
        self.test_mode = test_mode
        
        self.pipeline = Compose(pipeline)
        self.max_refetch = max_refetch

        if not lazy_init:
            self.full_init()

    
    def _slice_annotations(self, annotations, box_poly):
        cropped_annotations = []
        for annotation in annotations:
            seg = annotation['segmentation'][0]
            seg_poly = Polygon([[seg[i], seg[i+1]] for i in range(0, len(seg), 2)])
            if seg_poly.buffer(0).intersects(box_poly):
                cropped_seg_poly = seg_poly.buffer(0).intersection(box_poly)
                x1, y1, x2, y2 = map(int, list(cropped_seg_poly.bounds))
                x1 = x1 - box_poly.bounds[0]
                y1 = y1 - box_poly.bounds[1]
                x2 = x2 - box_poly.bounds[0]
                y2 = y2 - box_poly.bounds[1]

                if x2 > x1 and y2 > y1:
                    segmentation = []
                    if isinstance(cropped_seg_poly, MultiPolygon):
                        for poly in cropped_seg_poly.geoms:
                            for coord in list(poly.exterior.coords):
                                segmentation.append(coord[0] - box_poly.bounds[0])
                                segmentation.append(coord[1] - box_poly.bounds[1])
                    elif isinstance(cropped_seg_poly, Polygon):
                        for coord in list(cropped_seg_poly.exterior.coords):
                            segmentation.append(coord[0] - box_poly.bounds[0])
                            segmentation.append(coord[1] - box_poly.bounds[1])

                    if len(annotation["properties"][self.class_title_idx]["labels"]):
                        bbox_label = annotation["properties"][self.class_title_idx]["labels"][0]
                        assert bbox_label in range(len(self._metainfo["classes"]))

                        cropped_annotations.append({"area": cropped_seg_poly.area,
                                                    "bbox": [x1, y1, x2, y2],
                                                    "mask": [segmentation],
                                                    "bbox_label": bbox_label,
                                                    "ignore_flag": 0,
                                                    "extra_anns": []
                                                    })
        return cropped_annotations

    def _slice_image_info(self, vec_map, img_map):
        data_infos = []
        img_height = vec_map["images"][0]["height"]
        img_width = vec_map["images"][0]["width"]
        for slice_y in range(0, img_height, self.stride[0]):
            if (slice_y + self.window[0]) < img_height:
                for slice_x in range(0, img_width, self.stride[1]):
                    if (slice_x + self.window[1]) < img_width:
                        new_info = deepcopy(img_map)
                        new_info["slice_x"] = slice_x
                        new_info["slice_y"] = slice_y
                        new_info["slice_w"] = self.window[1]
                        new_info["slice_h"] = self.window[0]
                        box_poly = box(slice_x, slice_y, slice_x + self.window[1] - 1, slice_y + self.window[0] - 1)
                        new_info["instances"] = self._slice_annotations(vec_map["annotations"], box_poly)
                        data_infos.append(new_info)
                    elif slice_x < img_width and self.window[1] < img_width:
                        new_info = deepcopy(img_map)
                        new_info["slice_y"] = slice_y
                        slice_x = img_width - self.window[1] - 1
                        new_info["slice_x"] = slice_x
                        new_info["slice_w"] = self.window[1]
                        new_info["slice_h"] = self.window[0]
                        box_poly = box(slice_x, slice_y, slice_x + self.window[1] - 1, slice_y + self.window[0] - 1)
                        new_info["instances"] = self._slice_annotations(vec_map["annotations"], box_poly)
                        data_infos.append(new_info)
                    elif self.window[1] >= img_width:
                        new_info = deepcopy(img_map)
                        new_info["slice_y"] = slice_y
                        slice_x = 0
                        new_info["slice_x"] = slice_x
                        new_info["slice_w"] = img_width
                        new_info["slice_h"] = self.window[0]
                        box_poly = box(slice_x, slice_y, slice_x + img_width, slice_y + self.window[0] - 1)
                        new_info["instances"] = self._slice_annotations(vec_map["annotations"], box_poly)
                        data_infos.append(new_info)
   
            elif slice_y < img_height and self.window[0] < img_height:
                slice_y = img_height - self.window[0] - 1
                for slice_x in range(0, img_width, self.stride[1]):
                    if (slice_x + self.window[1]) < img_width:
                        new_info = deepcopy(img_map)
                        new_info["slice_x"] = slice_x
                        new_info["slice_y"] = slice_y
                        new_info["slice_w"] = self.window[1]
                        new_info["slice_h"] = self.window[0]
                        box_poly = box(slice_x, slice_y, slice_x + self.window[1] - 1, slice_y + self.window[0] - 1)
                        new_info["instances"] = self._slice_annotations(vec_map["annotations"], box_poly)
                        data_infos.append(new_info)
                    elif slice_x < img_width and self.window[1] < img_width:
                        new_info = deepcopy(img_map)
                        new_info["slice_y"] = slice_y
                        slice_x = img_width - self.window[1] - 1
                        new_info["slice_x"] = slice_x
                        new_info["slice_w"] = self.window[1]
                        new_info["slice_h"] = self.window[0]
                        box_poly = box(slice_x, slice_y, slice_x + self.window[1] - 1, slice_y + self.window[0] - 1)
                        new_info["instances"] = self._slice_annotations(vec_map["annotations"], box_poly)
                        data_infos.append(new_info)
                    elif self.window[1] >= img_width:
                        new_info = deepcopy(img_map)
                        new_info["slice_y"] = slice_y
                        slice_x = 0
                        new_info["slice_x"] = slice_x
                        new_info["slice_w"] = img_width
                        new_info["slice_h"] = self.window[0]
                        box_poly = box(slice_x, slice_y, slice_x + img_width, slice_y + self.window[0] - 1)
                        new_info["instances"] = self._slice_annotations(vec_map["annotations"], box_poly)
                        data_infos.append(new_info)

            elif self.window[0] >= img_height:
                slice_y = 0
                for slice_x in range(0, img_width, self.stride[1]):
                    if (slice_x + self.window[1]) < img_width:
                        new_info = deepcopy(img_map)
                        new_info["slice_x"] = slice_x
                        new_info["slice_y"] = slice_y
                        new_info["slice_w"] = self.window[1]
                        new_info["slice_h"] = img_height
                        box_poly = box(slice_x, slice_y, slice_x + self.window[1] - 1, slice_y + img_height)
                        new_info["instances"] = self._slice_annotations(vec_map["annotations"], box_poly)
                        data_infos.append(new_info)
                    elif slice_x < img_width and self.window[1] < img_width:
                        new_info = deepcopy(img_map)
                        new_info["slice_y"] = slice_y
                        slice_x = img_width - self.window[1] - 1
                        new_info["slice_x"] = slice_x
                        new_info["slice_w"] = self.window[1]
                        new_info["slice_h"] = img_height
                        box_poly = box(slice_x, slice_y, slice_x + self.window[1] - 1, slice_y + img_height)
                        new_info["instances"] = self._slice_annotations(vec_map["annotations"], box_poly)
                        data_infos.append(new_info)
                    elif self.window[1] >= img_width:
                        new_info = deepcopy(img_map)
                        new_info["slice_y"] = slice_y
                        slice_x = 0
                        new_info["slice_x"] = slice_x
                        new_info["slice_w"] = img_width
                        new_info["slice_h"] = img_height
                        box_poly = box(slice_x, slice_y, slice_x + img_width, slice_y + img_height)
                        new_info["instances"] = self._slice_annotations(vec_map["annotations"], box_poly)
                        data_infos.append(new_info)

        return data_infos

    def _load_vector_file(self, vec_path):
        with open(vec_path, 'r') as fin:
            vec_map = json.load(fin)

        file_name = vec_map["images"][0]["file_name"].split("/")[-1]
        file_path = os.path.join(self.raster_data_path, file_name)

        img_map = {
            "img_path": file_path,
            "img_id": vec_map["info"]["id"],
            "height": vec_map["images"][0]["height"],
            "width": vec_map["images"][0]["width"],
            "instances": [],
            "seg_fields": [],
            "reduce_zero_label": self.reduce_zero_label
        }

        if self.slice:
            return self._slice_image_info(vec_map, img_map)
        else:
            img_map["slice_x"] = 0
            img_map["slice_y"] = 0
            img_map["slice_w"] = img_map["width"]
            img_map["slice_h"] = img_map["height"]

            for annotation in vec_map["annotations"]:
                try:
                    xmin, ymin, width, height = annotation["bbox"]
                    xmax,ymax = xmin+width, ymin+height
                    bbox = [xmin, ymin, xmax, ymax]
                    
                    if len(annotation["properties"][self.class_title_idx]["labels"]):
                        bbox_label = annotation["properties"][self.class_title_idx]["labels"][0]

                        assert bbox_label in range(len(self._metainfo["classes"]))
                        
                        mask = [annotation["segmentation"][0][:-2]]

                        ann_map = {
                            "bbox": bbox,
                            "bbox_label": bbox_label,
                            "mask": mask,
                            "ignore_flag": 0,
                            "extra_anns": []
                        }

                        img_map["instances"] += [ann_map]
                except:
                    continue

            return img_map

    def load_data_list(self):
        meta_path = os.path.join(self.vector_data_path, "metadata.json")
        with open(meta_path, "r") as fin:
            metadata = json.load(fin)
    
        if not self.class_title:
            self.CLASSES = metadata["label:metadata"][0]["options"]
            self.class_title = metadata["label:metadata"][0]["title"].replace(" ", "-").lower()
            self.class_title_idx = 0
        else:
            for idx, question in enumerate(metadata["label:metadata"]):
                if question["title"] == self.class_title:
                    self.CLASSES = question["options"]
                    self.class_title = self.class_title.replace(" ", "-").lower()
                    self.class_title_idx = idx 

        self._metainfo = {
            "title": metadata["title"],
            "description": metadata["description"],
            "tags": metadata["tags"],
            "problemType": metadata["problemType"] if "problemType" in metadata else None,
            "question_title": metadata["label:metadata"][self.class_title_idx]["title"],
            "question_description": metadata["label:metadata"][self.class_title_idx]["description"],
            "classes": metadata["label:metadata"][self.class_title_idx]["options"],
            "reduce_zero_label": self.reduce_zero_label
        }

        vec_path_list = [os.path.join(self.vector_data_path, vec_file) for vec_file in metadata["dataset"][self.split]]

        pool = Pool(16)
        data_list = list(tqdm(pool.imap(self._load_vector_file, vec_path_list), total=len(vec_path_list)))
        if self.slice:
            data_list = [item for sublist in data_list for item in sublist]
        pool.close()

        tot_annotations = 0
        without_annotations = 0
        heights = []
        widths = []

        filtered_data_list = []
        for img in data_list:
            tot_annotation_this_img = len(img['instances'])
            if tot_annotation_this_img == 0:
                without_annotations += 1
            tot_annotations += tot_annotation_this_img
            heights.append(img['slice_h'])
            widths.append(img['slice_w'])
            if self.filter_empty and len(img['instances']):
                filtered_data_list.append(img)
            elif not self.filter_empty:
                filtered_data_list.append(img)
        
        print (f"Number of annotations in {self.split} : {tot_annotations}")
        print (f"Number of images without any annotations : {without_annotations}")
        print (f"Number of images used for training post filtering : {len(filtered_data_list)}")
        print (f"Minimum image size, Height : {min(heights)} Width : {min(widths)}")
        print (f"Maximum image size, Height : {max(heights)} Width : {max(widths)}")
        print (f"Average image size, Height : {np.mean(heights)} Width : {np.mean(widths)}")
    
        return filtered_data_list


from .enginelogger_hook import EngineLoggerHook
from .engine_cocoext import EngineCocoExt
from .engine_csv import EngineCSV
from .loading import LoadBandsFromFile, LoadVariableSizedBandsFromFile, LoadMasks
from .engine_cocometric import EngineCocoMetric
from .inference import Inference

__author__ = """Sagar Verma"""
__email__ = 'sagar@granular.ai'
__version__ = 'v0.0.17'

__all__ = ["EngineLoggerHook", "EngineCocoExt", "EngineCocoMetric",
           "EngineCSV", "LoadBandsFromFile", "LoadVariableSizedBandsFromFile",
           "LoadMasks", "Inference"]


import json, copy
import numpy as np
from tqdm import tqdm

import pandas as pd 

from mmengine.registry import DATASETS
from torch.utils.data import Dataset
from typing import List, Union, Callable

from mmengine.dataset.base_dataset import Compose
from multiprocessing import Pool


@DATASETS.register_module()
class EngineCSV(Dataset):
    def __init__(self,
                 data_path: str,
                 raster_dir_path: str,
                 vector_dir_path: str,
                 split: str = "train",
                 class_title: str = None,
                 pipeline: List[Union[dict, Callable]] = [],
                 max_refetch: int = 1000):
        self.raster_data_path = f"{data_path}/{raster_dir_path}"
        self.vector_data_path = f"{data_path}/{vector_dir_path}"

        self.split = split
        self.class_title = class_title

        self.pipeline = Compose(pipeline)
        self.max_refetch = max_refetch

        self.get_data_list()
    
    def full_init(self):
        self.get_data_info

    def process_record(self, record):

        file_name = record["image:01"].split("/")[-1]
        file_path = f"{self.raster_data_path}/{file_name}"

        return {
            "img_path": file_path,
            "img_id": record["image-id"],
            "gt_label": self.class_map[record[self.class_title]]
        }

    def get_data_list(self):
        meta_path = f"{self.vector_data_path}/metadata.json"
        metadata = json.load(open(meta_path, "r"))

        if not self.class_title:
            self.CLASSES = metadata["label:metadata"][0]["options"]
            self.class_title = metadata["label:metadata"][0]["title"].replace(" ", "-").lower()
            self.class_title_idx = 0
        else:
            for idx, question in enumerate(metadata["label:metadata"]):
                if question["title"] == self.class_title:
                    self.CLASSES = question["options"]
                    self.class_title = self.class_title.replace(" ", "-").lower()
                    self.class_title_idx = idx 

        self._metainfo = {
            "title": metadata["title"],
            "description": metadata["description"],
            "tags": metadata["tags"],
            "problemType": metadata["problemType"] if "problemType" in metadata else None,
            "question_title": metadata["label:metadata"][self.class_title_idx]["title"],
            "question_description": metadata["label:metadata"][self.class_title_idx]["description"],
            "classes": metadata["label:metadata"][self.class_title_idx]["options"]
        }

        self.class_map = {v:k for k,v in enumerate(metadata["label:metadata"][self.class_title_idx]["options"])}

        vec_path = f"{self.vector_data_path}/{metadata['dataset'][self.split]}"
        df = pd.read_csv(vec_path)
        records = df.to_dict('records')
        
        pool = Pool(16)
        self.data_list = list(tqdm(pool.imap(self.process_record, records), total=len(records)))
        pool.close()

    @property
    def metainfo(self) -> dict:
        """Get meta information of dataset.

        Returns:
            dict: meta information collected from ``BaseDataset.METAINFO``,
            annotation file and metainfo argument during instantiation.
        """
        return copy.deepcopy(self._metainfo)
    
    def get_data_info(self, idx: int):
        data = self.data_list[idx]
        data["sample_idx"] = idx

        return data
    
    def prepare_data(self, idx: int):
        data = self.get_data_info(idx)
        data['dataset'] = self
        return self.pipeline(data)

    def _rand_another(self):
        return np.random.randint(0, len(self))

    def __len__(self):
        return len(self.data_list)

    def __getitem__(self, idx: int):
        if self.split == "test":
            data = self.prepare_data(idx)
            if not data:
                raise Exception('Test time pipline should not get `None` data_sample')

        for _ in range(self.max_refetch+1):
            data = self.prepare_data(idx)
            if not data:
                idx = self._rand_another()
                continue

            return data



import torch
from typing import Dict, List, Optional, Sequence, Union

from mmdet.registry import METRICS
from mmdet.evaluation.metrics import CocoMetric
from mmdet.structures.mask import encode_mask_results


@METRICS.register_module()
class EngineCocoMetric(CocoMetric):
    def process(self, data_batch: dict, data_samples: Sequence[dict]) -> None:
        """Process one batch of data samples and predictions. The processed
        results should be stored in ``self.results``, which will be used to
        compute the metrics when all batches have been processed.

        Args:
            data_batch (dict): A batch of data from the dataloader.
            data_samples (Sequence[dict]): A batch of data samples that
                contain annotations and predictions.
        """
        for data_sample in data_samples:
            result = dict()
            pred = data_sample['pred_instances']
            result['img_id'] = data_sample['img_id']
            result['bboxes'] = pred['bboxes'].cpu().numpy()
            result['scores'] = pred['scores'].cpu().numpy()
            result['labels'] = pred['labels'].cpu().numpy()
            # encode mask to RLE
            if 'masks' in pred:
                result['masks'] = encode_mask_results(
                    pred['masks'].detach().cpu().numpy()) if isinstance(
                        pred['masks'], torch.Tensor) else pred['masks']
            # some detectors use different scores for bbox and mask
            if 'mask_scores' in pred:
                result['mask_scores'] = pred['mask_scores'].cpu().numpy()

            # parse gt
            gt = dict()
            gt['width'] = data_sample['ori_shape'][1]
            gt['height'] = data_sample['ori_shape'][0]
            gt['img_id'] = data_sample['img_id']

            if self._coco_api is None:
                # TODO: Need to refactor to support LoadAnnotations
                assert 'gt_instances' in data_sample, \
                    'ground truth is required for evaluation when ' \
                    '`ann_file` is not provided'
                gt['anns'] = [{
                    "bbox":data_sample['gt_instances']['bboxes'][idx].cpu().numpy(),
                    "bbox_label": int(data_sample['gt_instances']['labels'][idx].cpu().numpy())
                } for idx in range(len(data_sample['gt_instances']['labels']))]
            # add converted result to the results list
            self.results.append((gt, result))
